%{ open Ast %}

/* punctuation tokens */
%token LPAREN RPAREN LBRACE RBRACE LBRACK RBRACK SEMIC COMMA ASSIGN DOT
/* arithmentic operator tokens */
%token PLUS MINUS TIMES DIVIDE MOD
/* keyword tokens*/
%token DEF RETURN WHILE FOR NEW RHYTHM PITCH LENGTH IF ELSE
/* comparison tokens*/
%token EQ NEQ LT LEQ GT GEQ AND OR
/* data types tokens */
%token BOOL INT FLOAT NOTE STRING VOID
/* unary ops */
%token NOT

%token <string> PLIT ID FLIT STRLITERAL RLIT
%token <int> LITERAL
%token <bool> BLIT
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%left SEMIC
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT GEQ LEQ
%left PLUS MINUS
%left TIMES DIVIDE MOD
%right NOT

%start program
%type <Ast.program> program

%%

program:
  decls EOF { $1 }

decls:
   /* nothing */ { ([], [])               }
 | decls stmt { (($2 :: fst $1), snd $1) }
 | decls fdecl { (fst $1, ($2 :: snd $1)) }

fdecl:
   DEF typ ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE
     { { typ = $2;
	 fname = $3;
	 formals = List.rev $5;
	 body = List.rev $8 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { $1 }

formal_list:
    typ ID                   { [($1,$2)]     }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
   expr SEMIC { Expr($1) }
 | LBRACE stmt_list RBRACE                 { Block(List.rev $2)    }
 | RETURN expr_opt SEMIC                   { Return $2             }
 | typ ID SEMIC { Bind($1, $2) }
 | typ ID ASSIGN expr SEMIC { BindAssign($1, $2, $4)}
 | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
 | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7)        }
 | WHILE LPAREN expr RPAREN stmt { While ($3, $5) }
 | FOR LPAREN stmt expr SEMIC expr RPAREN stmt { For ($3,$4,$6, $8) }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          {$1 }

expr:
   PLIT { PLiteral($1) }
 | STRLITERAL { StrLiteral($1) }
 | LITERAL          { Literal($1) }
 | FLIT { Fliteral($1)}
 | BLIT { BoolLit($1)}
 | RLIT { RLiteral($1) }
 | accessible { $1 }
 | ID LPAREN args_opt RPAREN { Call($1, $3) }
  | accessible ASSIGN expr {Assign ($1, $3) }
  /* math */
  | expr PLUS   expr { Binop($1, Add, $3) }
  | expr MINUS  expr { Binop($1, Sub, $3) }
  | expr TIMES  expr { Binop($1, Mult, $3) }
  | expr DIVIDE expr { Binop($1, Div, $3) }
  | expr MOD    expr { Binop($1, Mod, $3) }
  | LPAREN expr RPAREN { $2 }
  /* logic */
  | expr EQ     expr { Binop($1, Equal, $3)   }
  | expr NEQ    expr { Binop($1, Neq,   $3)   }
  | expr LT     expr { Binop($1, Less,  $3)   }
  | expr LEQ    expr { Binop($1, Leq,   $3)   }
  | expr GT     expr { Binop($1, Greater, $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3)   }
  | expr AND    expr { Binop($1, And,   $3)   }
  | expr OR     expr { Binop($1, Or,    $3)   }
  /* Unary */
  | MINUS expr %prec NOT { Unop(Neg, $2)      }
  | NOT expr         { Unop(Not, $2)}
  /* note constructor */
  | NEW NOTE LPAREN args_opt RPAREN { CreateNote($4) }
  /* note access */
  | accessible DOT RHYTHM { NoteAccess($1, R) }
  | accessible DOT PITCH  { NoteAccess($1, P) }
  | accessible DOT RHYTHM ASSIGN expr { NoteAssign($1, R, $5) }
  | accessible DOT PITCH ASSIGN expr { NoteAssign($1, P, $5) }
 /* array constructor and accessor */
  | NEW typ LBRACK expr RBRACK { CreateArray($2, $4) }
  | LBRACK args_opt RBRACK { ArrayLiteral($2) }
  | accessible DOT LENGTH { ArrayLength($1) }

accessible:
    ID { Id($1) }
  | accessible LBRACK expr RBRACK { ArrayAccess($1, $3) }

args_opt:
    /* nothing */ { [] }
  | args_list  { List.rev $1 }

args_list:
    expr                    { [$1] }
  | args_list COMMA expr { $3 :: $1 }

typ:
    VOID  { Void }
  | INT { Int }
  | BOOL { Bool }
  | NOTE { Note }
  | FLOAT { Float }
  | STRING { String }
  | typ LBRACK RBRACK { Array($1)}
